/*
    Filename: x.c
    Synopsis: X framework.

                          Copyright (C) 2000-2004
            Ral Nez de Arenas Coronado <raul@pleyades.net>
	         David Gmez Espinosa <david@pleyades.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA 02111-1307 USA

*/

#include "config.h"
#include "x.h"

#define BUFSIZE 1024
#define PIXMAPDIR "/usr/share/pixmaps"
#define MIN_PIXMAPW 16
#define MIN_PIXMAPH 16
#define MAX_PIXMAPW 64
#define MAX_PIXMAPH 64


/*
    NOTE: the pointer is declared 'static' here and not loaded from a file in
    order to obtain the 'blackbox' effect. This way we have all resources
    encapsulated in the binary file obtained.
*/

/* This is the default pointer: a pretty arrow */
#define CURSORW 14
#define CURSORH 14
#define CURSORX 1
#define CURSORY 1

static unsigned char arrow_bits[] = {
   0x00, 0x00, 0x06, 0x00, 0x1e, 0x00, 0x7c,
   0x00, 0xfc, 0x01, 0xf8, 0x07, 0xf8, 0x1f,
   0xf0, 0x01, 0xf0, 0x03, 0x60, 0x07, 0x60,
   0x0e, 0x40, 0x1c, 0x40, 0x18, 0x00, 0x00};

static unsigned char arrowmask_bits[] = {
   0x07, 0x00, 0x1f, 0x00, 0x7f, 0x00, 0xfe,
   0x01, 0xfe, 0x07, 0xfc, 0x1f, 0xfc, 0x3f,
   0xf8, 0x3f, 0xf8, 0x07, 0xf0, 0x0f, 0xf0,
   0x1f, 0xe0, 0x3e, 0xe0, 0x3c, 0xc0, 0x38};


/* Global resources commonly used */
GC xgc=None;
Display *display=NULL;
char *display_name=NULL;
Cursor x_root_cursor=None;

/*
    This struct defines a widget callback.
    These callbacks are called whenever an event arrives whose destination
is the window in the Sloot, and is called using the 'widget' as parameter.
*/
typedef struct {
    Window window;
    Handler handle;
    void *widget;
} Sloot;

static Chain sloots;  /* Stores the Sloots */

/* Our handlers for X errors */
static int _errorhandler   (Display *, XErrorEvent *);
static int _ioerrorhandler (Display *);

static unsigned short int reference_count=0;





/* This function initializes our fantastic X framework ;) */
void x_init (void) {

    XVisualInfo vinfo;

    if (reference_count++) return;

    /* First we connect to the X server */
    /* Try to open default display */
    display = XOpenDisplay(NULL);
    if (!display) die("Pawm is unable to open display");
    
    /* Deactivate screen saver by now. We won't use extensions. */
    XSetScreenSaver(display, 0, 0, DefaultBlanking, DefaultExposures);

    XStoreName(display, ROOT, "root");

    /* The default GC */        
    xgc=XCreateGC(display, ROOT, 0, NULL);
    if (xgc==None) BUG();

    /* Setup colors */
    if (!XMatchVisualInfo(display, SCREENI, DEPTH, TrueColor, &vinfo)) BUG();

    x_root_cursor=x_build_cursor(arrow_bits, arrowmask_bits,
                                 CURSORW, CURSORH, CURSORX, CURSORY);
    
    //x_root_cursor=XCreateFontCursor(display, XC_arrow);
    XDefineCursor(display, ROOT, x_root_cursor);

    /* Install the error handlers */
    XSetErrorHandler(_errorhandler);
    XSetIOErrorHandler(_ioerrorhandler);

    /* Create the Sloot list */
    sloots=chain_create();
    if (!sloots) BUG();

    /* ICCCM atoms */
    atoms.string=XInternAtom(display, "STRING", False);
    atoms.wm_name=XInternAtom(display, "WM_NAME", False);
    atoms.wm_icon_name=XInternAtom(display, "WM_ICON_NAME", False);
    atoms.wm_normal_hints=XInternAtom(display, "WM_NORMAL_HINTS", False);
    atoms.wm_transient_for=XInternAtom(display, "WM_TRANSIENT_FOR", False);
    atoms.wm_state=XInternAtom(display, "WM_STATE", False);
    atoms.wm_delete_window=XInternAtom(display, "WM_DELETE_WINDOW", False);
    atoms.wm_protocols=XInternAtom(display, "WM_PROTOCOLS", False);
    atoms.wm_change_state=XInternAtom(display, "WM_CHANGE_STATE", False);
    /* EWMH */
    atoms.utf8_string=XInternAtom(display, "UTF8_STRING", False);
    atoms.net_wm_name=XInternAtom(display, "_NET_WM_NAME", False);
    atoms.net_wm_icon_name=XInternAtom(display, "_NET_WM_ICON_NAME", False);
    atoms.net_wm_state=XInternAtom(display, "_NET_WM_STATE", False);
    atoms.net_wm_state_modal=XInternAtom(display, "_NET_WM_STATE_MODAL", False);
    atoms.net_wm_state_sticky=XInternAtom(display, "_NET_WM_STATE_STICKY", False);
    atoms.net_wm_state_maximized=XInternAtom(display, "_NET_WM_STATE_MAXIMIZED", False);
    atoms.net_wm_state_shaded=XInternAtom(display, "_NET_WM_STATE_SHADED", False);
    atoms.net_wm_state_skip_taskbar=XInternAtom(display, "_NET_WM_STATE_SKIP_TASKBAR", False);
    atoms.net_wm_state_skip_pager=XInternAtom(display, "_NET_WM_SKIP_PAGER", False);
    atoms.net_wm_state_hidden=XInternAtom(display, "_NET_WM_STATE_HIDDEN", False);
    atoms.net_wm_state_fullscreen=XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", False);
    atoms.net_wm_state_above=XInternAtom(display, "_NET_WM_STATE_ABOVE", False);
    atoms.net_wm_state_below=XInternAtom(display, "_NET_WM_STATE_BELOW", False);
    atoms.net_supported=XInternAtom(display, "_NET_SUPPORTED", False);
    atoms.net_current_desktop=XInternAtom(display, "_NET_CURRENT_DESKTOP", False);
    atoms.net_active_window=XInternAtom(display, "_NET_ACTIVE_WINDOW", False);
    atoms.net_workarea=XInternAtom(display, "_NET_WORKAREA", False);
    atoms.net_supporting_wm_check=XInternAtom(display, "_NET_SUPPORTING_WM_CHECK", False);
    atoms.net_number_of_desktops=XInternAtom(display, "_NET_NUMBER_OF_DESKTOPS", False);
    atoms.net_client_list=XInternAtom(display, "_NET_CLIENT_LIST", False);
    /* Other atoms */
    atoms.motif_wm_hints=XInternAtom(display,"_MOTIF_WM_HINTS", False);

}




/* This function shuts down our fantastic X framework :((((*/
void x_done (void) {

    Sloot *victim=NULL;
    
    if (--reference_count) return;
    
    /* Let our sloots run away free ;), they did a good work */
    chain_seek(sloots, chain_size(sloots));
    while (chain_size(sloots)) {
        victim=chain_del(sloots);
	free(victim);
    }
    
    /* Destroy the chain itself */
    chain_delete(sloots);

    /* Free resources... */
    XFreeGC(display, xgc);
    XFreeCursor(display, x_root_cursor);

    XCloseDisplay(display);
    
}




/* This function returns a pixel value from a color name which value
 * is specified in the pawm config file */
unsigned long x_color(const char *color) {
	
    XColor scolor,ecolor;
    char *tmpvalue;

    tmpvalue=cfgpool_getstring(pawmpool,color);
    /* White as default if not color is found */
    if (tmpvalue==NULL) { 
	SAY("Color %s not found in config file",color);
	return WHITE;
    }
    XAllocNamedColor(display,COLORMAP,tmpvalue,&scolor,&ecolor);
    free(tmpvalue);

    return scolor.pixel;
}


char *x_colorstr(const char *color) {
	
    char *tmpvalue;

    tmpvalue=cfgpool_getstring(pawmpool,color);
    /* White as default if not color is found */
    if (tmpvalue==NULL) { 
	SAY("Color %s not found in config file",color);
	return "white";
    }

    return tmpvalue;
}


/* This function returns a pixel value from an RGB specification */
unsigned long x_pixel(const char *pixel) {

    XColor scolor,ecolor;

    XAllocNamedColor(display,COLORMAP,pixel, &scolor,&ecolor);

    return scolor.pixel;
}


/*
    This builds a Cursor from bitmap data.
    The caller must free the returned Cursor.
    The function returns 'None' on error.
*/
Cursor x_build_cursor (unsigned char *shape_bits,
                       unsigned char *mask_bits,
                       unsigned int w, unsigned int h,
                       int x, int y) {

    Pixmap shape=None;
    Pixmap mask=None;
    XColor fore;
    XColor back;
    Cursor cursor;

    /* Build the pointer */
    shape=XCreatePixmapFromBitmapData(display, ROOT, shape_bits,
                                      w, h, 1, 0, 1);
    mask =XCreatePixmapFromBitmapData(display, ROOT, mask_bits,
                                      w, h, 1, 0, 1);

    if ((shape==None) || (mask==None)) BUG();

    fore.pixel=WHITE;
    back.pixel=BLACK;
    if (!XQueryColor(display, COLORMAP, &fore)) BUG();
    if (!XQueryColor(display, COLORMAP, &back)) BUG();

    cursor=XCreatePixmapCursor(display, shape, mask, &fore, &back,
	                      (unsigned int)x,(unsigned int) y);
    if (cursor==None) BUG();

    XFreePixmap(display, shape);
    XFreePixmap(display, mask);
        
    return cursor;
}



/* This function creates a pixmap from a icon in xpm format, also returning
 * an icon mask for XPM's with transparency. All the data is contained in
 * a xImage type */
xImage x_build_pixmap(char *xpmicon)
{
    XpmAttributes attr;
    char xpmpath[BUFSIZE];
    struct stat sinfo;
    int rvalue;
    xImage image;
    char *end_xpmpath;

    ASSERT(xpmicon);

    snprintf(xpmpath, BUFSIZE, "%s/icons/%s", DATADIR, xpmicon);
    if ( stat(xpmpath, &sinfo)==-1 ) {
        snprintf(xpmpath, BUFSIZE, "%s/%s", PIXMAPDIR, xpmicon);
	if (stat(xpmpath, &sinfo)==-1) return NULL;
    }

    image=malloc(sizeof(struct sxImage));
    image->pixmap=None; image->mask=None;

    /* Use default values */
    attr.valuemask=0;
    rvalue=XpmReadPixmapFile(display,ROOT,xpmpath,
	                     &(image->pixmap),&(image->mask),&attr);    
    if (rvalue!=XpmSuccess) {
	/* Fallback to 256 color icons...*/
	SAY("Using 256 color icons...");
	
        end_xpmpath=xpmpath+strlen(xpmpath)-4;
	strcat(end_xpmpath,"_256.xpm");
        rvalue=XpmReadPixmapFile(display,ROOT,xpmpath,
	                         &(image->pixmap),&(image->mask),&attr);    
	if (rvalue!=XpmSuccess) {
	    bang("XPM cannot be created");
	    free(image);
            return NULL;
	}    
    }

    /* Check size limits for XPM files. They're reasonable limits for
       all resolutions, although they can be changed without problems */
    if ( (attr.width < MIN_PIXMAPW) || (attr.width > MAX_PIXMAPW) ||
         (attr.height < MIN_PIXMAPH) || (attr.height > MAX_PIXMAPH)) {
        fprintf(stderr, "%s xpm size is too big or too small", xpmpath);
    }

    image->w=attr.width; image->h=attr.height;

    XpmFreeAttributes(&attr);

    return image;
    
}



void x_destroy_pixmap(xImage icon)
{
    if (icon->pixmap!=None) XFreePixmap(display, icon->pixmap);
    if (icon->mask!=None) XFreePixmap(display, icon->mask);

    free(icon);
}
    


/*
    This function returns the number of full chars of 'text' which
will fit in the given width, 'w', if painted with 'font'.

NOTE: uses Bolzano's theorem to calculate the size.
*/
size_t x_trimtext (XFontStruct *font, const char *text, unsigned int w) {

    size_t min=0;
    size_t max=0;
    size_t pivot=0;

    ASSERT(font);
    if (!text) return 0;

    max=strlen(text)+1;
    pivot=max/2;    

    /* We use Bolzano's theorem for calculate the text width */
    while (1) {
        unsigned int pivotw=0;

        /* Compute text length in pixels for 'pivot' chars */
        pivotw=XTextWidth(font, text, (int)pivot);
        
        /* If the minimum is greater than the allowed width... */
        if (pivotw >= w) max=pivot;
        if (pivotw < w) min=pivot;
        if ((max-min) == 1) return min;  /* Algorithm converged */
        pivot=((max-min)/2)+min;
    }

}

/* Get data associated to an X atom */
unsigned char *x_get_property(Window window, Atom atom, Atom type,
                	      unsigned long *nitems)
{
    unsigned char *data;
    Atom rettype;
    int retformat;
    unsigned long nbytes;
    XTextProperty text;

    /* First look for known atoms, if not get the raw property data, the
       caller is responsible to know what is asking for */
    switch(atom) {
	case XA_WM_NAME:
	    if (XGetWMName(display, window, &text)) {;
	        data=text.value;
		*nitems=strlen(text.value);
            } else {
		data=NULL;
                SAY("No WM_NAME found");		
	    }	
	    break;
	case XA_WM_ICON_NAME:
	    if (XGetWMIconName(display, window, &text)) {;
	        data=text.value;
		*nitems=strlen(text.value);
            } else {
		data=NULL;
                SAY("No WM_ICON_NAME found");		
	    }	
	    break;
	default:    
            if (XGetWindowProperty(display, window, atom, 0L, 40L, False,
	    		           type, &rettype, &retformat,
		                   nitems, &nbytes, &data)!=Success) return NULL;
            if (rettype==None) {
		SAY("No %s atom found", XGetAtomName(display, atom));
		data=NULL;
	    } 
    }				   

    return data;			   
}


void x_set_property(Window window, Atom atom, Atom type, unsigned char *data, int nitems)
{
    
      SAY("Setting X property %s",XGetAtomName(display, atom));

      XChangeProperty(display, window, atom, type, 32, PropModeReplace,
		      data, nitems);
}

void x_set_property_utf8(Window window, Atom atom, unsigned char *data)
{
    
      SAY("Setting X utf8 property %s",XGetAtomName(display, atom));

      XChangeProperty(display, window, atom, atoms.utf8_string, 8, PropModeReplace,
		      data, strlen(data));
}


/*
    QT names the event handlers 'slots'.
    We use sloot, because we are the Puto Amo.
*/
void x_addsloot (Window window, Handler handler, void *widget) {

    Sloot *sloot=NULL;

    ASSERT(window!=None);
    ASSERT(handler);
    ASSERT(sloots);

    /* Create the new Sloot */
    sloot=malloc(sizeof(Sloot));
    if (!sloot) die("Not enough memory");

    sloot->window=window;
    sloot->handle=handler;
    sloot->widget=widget;

    /* We can safely add duplicates :(( */
    chain_seek(sloots, chain_size(sloots));
    chain_put(sloots, sloot);
    

}




/* Remember: we are the Puto Amo. */
void x_delsloot (Window window, Handler handler, void *widget) {

    Sloot *badboy;

    ASSERT(window!=None);
    ASSERT(handler);
    ASSERT(sloots);

    chain_home(sloots);

    while ( (badboy=chain_next(sloots))!=NULL ) {

     if ((badboy->window==window) && (badboy->handle==handler) &&
	 (badboy->widget=widget)) {
	chain_del(sloots);
	free(badboy);
	return;
     }
    }
	
}


/*
Non fatal error. We will use it for dealing with a race condition:
sometimes a window is destroyed while handling some event, and we
don't notice it, so it generates BadWindow and the like...
*/
static int _errorhandler (Display *dpy, XErrorEvent *event) {

    char buffer[80];

    /* Assert the values */
    ASSERT(dpy);
    ASSERT(dpy==display);
    ASSERT(dpy==event->display);

    dpy=NULL;

    XGetErrorText(display,event->error_code,buffer,sizeof(buffer));

    /* Always emit this info. Helps debugging. */
    SAY("Got non fatal error on XID %#lx:\n"\
        "\tSerial number: %lu\n"\
        "\tError message: %s [%u]\n"\
        "\tRequest code: %u [%u]\n",
        event->resourceid,
        event->serial,  
        buffer, event->error_code,
        event->request_code, event->minor_code);

     /*
        Trying to debug an X app using the error handlers is just
     like smashing your own dick with a really big hammer: it may look
     interesting, but after a few tries you notice that is not a very
     good idea...
     */

     return 0;
}




/* Fatal (lost connections and the like) errors */
static int _ioerrorhandler (Display *dpy) {

    ASSERT(dpy==display);

    /* Por decir algo... */
    die("Fatal IO error on display %s, goodbye :))", DisplayString(dpy));

    return 0;  /* Make gcc happy */
}




/*
    This function returns the next event from the queue.
    It blocks until a new event arrives.
    It returns the event in a static storage which is overwritten in
future calls, so it is NON-REENTRANT. Be careful. Caveat emptor.

NOTE: This function may change soon in the future.
*/
XEvent *x_getevent (void) {
    static XEvent event;

    XNextEvent(display,&event);	    
    
    return (&event);
}




/* Dispatch one event. Please note: just ONE EVENT. */
void x_dispatch (XEvent *xevent) {

    Window window=None;
    Sloot *sloot=NULL;
    size_t oldpos;

    /* ...determine destination... */
    window=xevent->xany.window;

    if (window==None) return;

    /* Before messing with the sloots chain, save the cursor offset. This
       way we can restore it later, and allow nested x_dispatch calls */
    oldpos=chain_offset(sloots);

    /* ...and run the aproppriate handler, if any. */
    chain_home(sloots);
    while ( (sloot=chain_next(sloots))!=NULL ) {
        if (sloot->window==window) {
	    sloot->handle(xevent, sloot->widget);
	}
    }

    chain_seek(sloots, oldpos);
    
}
